! Likelihood for Dutcher et al. 2020
! SPT-3G Y1 95, 150, 220GHz bandpowers, l=300-3000, EE/TE
! Written by Lennart Balkenhol

! N_b = number of bins
! N_s = number of spectra
! N_be = number of beam error parameters
! N_f = number of fields
! Subscript _0 denotes uncropped parameters

module CMB_SPT3G_EETE_2020
use Likelihood_Cosmology
use CosmoTheory
use CosmologyTypes
use FileUtils
use MatrixUtils
use settings

implicit none

! Declare global variables
integer :: SPT3G_windows_lmin, SPT3G_windows_lmax

Type(TStringList) :: full_bandpower_list ! List specifying the order of bandpowers in the bandpower file
Type(TStringList) :: full_window_list ! List specifying the order of bandpowers in the window files
Type(TStringList) :: full_covariance_list ! List specifying the order of spectra in the covariance matrix file
Type(TStringList) :: full_beam_covariance_list ! List specifying the order of spectra in the beam covariance matrix file
Type(TStringList) :: nu_eff_list ! List specifying the order of spectra in the central frequency file
Type(TStringList) :: full_beam_error_list ! List specifying the order of spectra in the beam error file
Type(TStringList) :: spectra_to_fit_list ! List specifying which spectra to use for parameter estimation

integer, allocatable :: spectra_to_fit_window_indices(:) ! Gives the index in the window files of each spectrum to be used
integer, allocatable :: spectra_to_fit_bandpower_indices(:) ! Gives the index in the bandpower file of each spectrum to be used
integer, allocatable :: spectra_to_fit_nu_eff_indices(:) ! Gives the index in the central frequency file of each spectrum to be used
integer, allocatable :: spectra_to_fit_poisson_indices(:) ! Gives the indices in the DataParams object for the poisson amplitude
integer, allocatable :: spectra_to_fit_tcal_indices(:,:) ! Gives the indices in the DataParams object for the TT spec calibration
integer, allocatable :: spectra_to_fit_ecal_indices(:,:) ! Gives the indices in the DataParams object for the EE spec calibration
integer, dimension(6) :: cal_rows_use ! Convenience array indicating whether cal 90/150/220 is used

real(mcp), parameter :: T_CMB = 2.72548_mcp    ! CMB temperature
real(mcp), parameter :: h = 6.62606957e-34_mcp ! Planck's constant
real(mcp), parameter :: kB = 1.3806488e-23_mcp ! Boltzmann constant
real(mcp), parameter :: Ghz_Kelvin = h/kB*1e9_mcp

! Level of feedback
logical :: print_spectrum

! Reference frequencies for foregrounds that need to be scaled (GHz)
real(mcp) :: nu_0_radio ! Unresolved radio galaxies
real(mcp) :: nu_0_dust ! Polarised galactic dust
real(mcp) :: nu_0_dsfg ! Dusty star-forming galaxies

integer, parameter :: N_freq_0 = 3 ! Number of frequency bands
integer, parameter :: N_b_0 = 44 ! Maximum (uncropped) number of bandpowers
integer, parameter :: N_s_0 = 12 ! Maximum (uncropped) number of spectra that can be used for parameter estimation

! Simple foreground model:
! Use individual amplitudes for EE poisson foregrounds and forget about frequency scaling, zero TE
logical :: use_simple_poisson_foregrounds
! Order of the data parameters in the .paramnames file
! \kappa (super sample lensing)
! Poisson sources
! --- The remaining indices are assigned later when the foreground model is read in
! Polarised galactic dust
! Calibration factors
! First beam factor
integer :: iKappa=1,&
  iPoisson_90x90=2, iPoisson_90x150=3, iPoisson_90x220=4, iPoisson_150x150=5, iPoisson_150x220=6, iPoisson_220x220=7,&
  iTDust=8, iADust_TE_150=9, iBetaDust_TE=10, iAlphaDust_TE=11, iADust_EE_150=12,&
  iBetaDust_EE=13, iAlphaDust_EE=14,&
  iMapTcal90=15, iMapTcal150=16, iMapTcal220=17,&
  iMapEcal90=18, iMapEcal150=19, iMapEcal220=20

! Indices from old foreground model
! Remaining adjustments are made when paramfilename is read in according to model specified
integer :: iDl_Radio_TE_150=2, iAlphaRadio_TE=3, iDl_Radio_EE_150=4, iAlphaRadio_EE=5,&
  iTDSFG=6, iDl_DSFG_TE_150=7, iBetaDSFG_TE=8, iDl_DSFG_EE_150=9, iBetaDSFG_EE=10

! Nuisance parameter priors
real(mcp) :: Kappa_prior_mean, Kappa_prior_sigma
real(mcp) :: AlphaDustEE_prior_mean, AlphaDustEE_prior_sigma
real(mcp) :: AlphaDustTE_prior_mean, AlphaDustTE_prior_sigma
real(mcp) :: BetaDustEE_prior_mean, BetaDustEE_prior_sigma
real(mcp) :: BetaDustTE_prior_mean, BetaDustTE_prior_sigma
real(mcp) :: BeamCovScaling
integer :: Cal_prior_switch, Kappa_prior_switch, AlphaDustEE_prior_switch, AlphaDustTE_prior_switch,&
  BetaDustEE_prior_switch, BetaDustTE_prior_switch

Type, extends(TCMBLikelihood) :: TSPT3G_EETE_Likelihood
  integer :: N_s, N_b, N_freq, bin_min, bin_max
  integer :: poisson_switch, ssl_switch, radio_gal_switch, dsfg_switch, dust_switch
  real(mcp) :: aberration_coefficient
  real(mcp), allocatable :: full_bandpowers(:,:)
  real(mcp), allocatable :: bdp_covariance(:,:)
  real(mcp), allocatable :: beam_covariance(:,:)
  real(mcp), allocatable :: cal_inv_covariance(:,:)
  real(mcp), allocatable :: full_windows(:,:,:)
  real(mcp), allocatable :: spectra_to_fit_nu_eff(:,:)
contains
  procedure :: ReadIni => SPT3G_EETE_ReadIni
  procedure :: SPT3G_EETE_InitData
  procedure :: LogLike => SPT3G_EETE_LogLike 
end Type TSPT3G_EETE_Likelihood

contains

subroutine SPT3G_EETE_ReadIni(this, Ini)
  class(TSPT3G_EETE_Likelihood) :: this
  class(TSettingIni) :: Ini

  character(LEN=:), allocatable :: parameter_filename
  
  character(LEN=:), allocatable :: bandpower_filename
  character(LEN=:), allocatable :: full_bandpower_list_string

  character(LEN=:), allocatable :: covariance_filename
  character(LEN=:), allocatable :: full_covariance_list_string

  character(LEN=:), allocatable :: beam_covariance_filename
  character(LEN=:), allocatable :: full_beam_covariance_list_string

  character(LEN=:), allocatable :: cal_covariance_filename

  character(LEN=:), allocatable :: window_folder
  character(LEN=:), allocatable :: full_window_list_string
  
  character(LEN=:), allocatable :: nu_eff_filename
  character(LEN=:), allocatable :: nu_eff_list_string

  character(LEN=:), allocatable :: spectra_to_fit_list_string

  print *, "SPT-3G Y1 EE/TE: Requested Likelihood. Beginning initialisation."

  SPT3G_windows_lmin = Ini%Read_Int("SPT3G_Y1_EETE_window_l_min")
  SPT3G_windows_lmax = Ini%Read_Int("SPT3G_Y1_EETE_window_l_max")

  allocate(this%cl_lmax(CL_E,CL_E), source=0)
  this%cl_lmax(CL_T,CL_T) = SPT3G_windows_lmax+1
  this%cl_lmax(CL_E,CL_T) = SPT3G_windows_lmax+1
  this%cl_lmax(CL_E,CL_E) = SPT3G_windows_lmax+1

  ! Read in aberration correction coefficient
  this%aberration_coefficient = Ini%Read_Real("SPT3G_Y1_EETE_aberration_coefficient", 0.0)

  ! Read in foreground model
  use_simple_poisson_foregrounds = Ini%Read_Logical("SPT3G_Y1_EETE_simple_poisson_foregrounds", .true.)

  ! Read in switches for foregrounds
  this%ssl_switch = LogicalToInt( Ini%Read_Logical("SPT3G_Y1_EETE_super_sample_lensing", .true.) )
  this%poisson_switch = LogicalToInt( Ini%Read_Logical("SPT3G_Y1_EETE_poisson_sources", .true.) )
  this%dust_switch = LogicalToInt( Ini%Read_Logical("SPT3G_Y1_EETE_dust", .true.) )

  ! Read in foreground reference frequencies
  nu_0_radio = Ini%Read_Real("SPT3G_Y1_EETE_radio_galaxies_nu0", 150.0)
  nu_0_dsfg = Ini%Read_Real("SPT3G_Y1_EETE_dsfg_nu0", 150.0)
  nu_0_dust = Ini%Read_Real("SPT3G_Y1_EETE_dust_nu0", 150.0)

  ! Read in spectrum selection
  spectra_to_fit_list_string = Ini%Read_String("SPT3G_Y1_EETE_spectra_to_fit", .true.)
  call spectra_to_fit_list%SetFromString(spectra_to_fit_list_string)
  this%N_s = spectra_to_fit_list%Count

  ! Read in bin selection
  this%bin_min = Ini%Read_Int("SPT3G_Y1_EETE_bin_min", 1, min=1, max=N_b_0)
  this%bin_max = Ini%Read_Int("SPT3G_Y1_EETE_bin_max", N_b_0, min=this%bin_min, max=N_b_0)
  this%N_b = this%bin_max-this%bin_min+1

  ! Read in nuisance parameter priors
  Cal_prior_switch = LogicalToInt( Ini%Read_Logical("SPT3G_Y1_EETE_cal_prior", .false.) )

  Kappa_prior_switch = LogicalToInt( Ini%Read_Logical("SPT3G_Y1_EETE_kappa_prior", .false.) )
  Kappa_prior_mean = Ini%Read_Real("SPT3G_Y1_EETE_Kappa_mean", 0.0)
  Kappa_prior_sigma = Ini%Read_Real("SPT3G_Y1_EETE_Kappa_sigma", 0.00045)

  AlphaDustEE_prior_switch = LogicalToInt( Ini%Read_Logical("SPT3G_Y1_EETE_Alpha_Dust_EE_prior", .false.) )
  AlphaDustEE_prior_mean = Ini%Read_Real("SPT3G_Y1_EETE_Alpha_Dust_EE_prior_mean", -2.42)
  AlphaDustEE_prior_sigma = Ini%Read_Real("SPT3G_Y1_EETE_Alpha_Dust_EE_prior_sigma", 0.02)

  AlphaDustTE_prior_switch = LogicalToInt( Ini%Read_Logical("SPT3G_Y1_EETE_Alpha_Dust_TE_prior", .false.) )
  AlphaDustTE_prior_mean = Ini%Read_Real("SPT3G_Y1_EETE_Alpha_Dust_TE_prior_mean", -2.42)
  AlphaDustTE_prior_sigma = Ini%Read_Real("SPT3G_Y1_EETE_Alpha_Dust_TE_prior_sigma", 0.02)

  BeamCovScaling = Ini%Read_Real("SPT3G_Y1_EETE_Beam_Covariance_Scaling", 1.0)
  ! Read in file locations (parameter file, band powers, covariance, window functions, effective frequencies, beam error)
  if (use_simple_poisson_foregrounds) then
      parameter_filename = Ini%Read_String_Default("SPT3G_Y1_EETE_params_file","")
  else
      parameter_filename = Ini%Read_String_Default("SPT3G_Y1_EETE_params_file_old_foregrounds","")
      ! Set parameter indices that occur in both foreground models, deprecated!
      iTDust=11; iADust_TE_150=12; iBetaDust_TE=13; iAlphaDust_TE=14; iADust_EE_150=15;&
        iBetaDust_EE=16; iAlphaDust_EE=17;
  end if
  call this%loadParamNames(parameter_filename)

  bandpower_filename = Ini%ReadFileName("SPT3G_Y1_EETE_bandpower_file",relative = .true., NotFoundFail=.true.)
  full_bandpower_list_string = Ini%Read_String("SPT3G_Y1_EETE_bandpower_file_order", .true.)
  call full_bandpower_list%SetFromString(full_bandpower_list_string)

  covariance_filename = Ini%ReadFileName("SPT3G_Y1_EETE_covariance_matrix",relative = .true., NotFoundFail=.true.)
  full_covariance_list_string = Ini%Read_String("SPT3G_Y1_EETE_covariance_matrix_order", .true.)
  call full_covariance_list%SetFromString(full_covariance_list_string)

  beam_covariance_filename = Ini%ReadFileName("SPT3G_Y1_EETE_beam_covariance_matrix",relative = .true., NotFoundFail=.true.)
  full_beam_covariance_list_string = Ini%Read_String("SPT3G_Y1_EETE_beam_covariance_matrix_order", .true.)
  call full_beam_covariance_list%SetFromString(full_beam_covariance_list_string)

  cal_covariance_filename = Ini%ReadFileName("SPT3G_Y1_EETE_cal_covariance_matrix",relative = .true., NotFoundFail=.true.)

  window_folder = Ini%Read_String("SPT3G_Y1_EETE_window_folder", NotFoundFail=.true.)
  full_window_list_string = Ini%Read_String("SPT3G_Y1_EETE_window_folder_order", .true.)
  call full_window_list%SetFromString(full_window_list_string)

  nu_eff_filename = Ini%ReadFileName("SPT3G_Y1_EETE_central_frequency_file",relative = .true., NotFoundFail=.true.)
  nu_eff_list_string = Ini%Read_String("SPT3G_Y1_EETE_central_frequency_file_order", .true.)
  call nu_eff_list%SetFromString(nu_eff_list_string)

  ! Check level of feedback
  print_spectrum = Ini%Read_Logical('print_spectrum',.false.)
  if (print_spectrum .and. MPIRank /= 0) then
    call MPIStop("SPT-3G Y1 EE/TE: print_spectrum is not MPI thread-safe!")
  endif

  print *, "USING BANDPOWER FILE: ", bandpower_filename

  ! Announce prior selection
  print *, "SPT-3G Y1 EE/TE: Requested priors:"

  print *, "Cal prior: ", IntToLogical(cal_prior_switch)

  print *, "Kappa prior: ", IntToLogical(Kappa_prior_switch)
  print *, "Kappa mean: ", Kappa_prior_mean
  print *, "Kappa prior: ", Kappa_prior_sigma

  print *, "AlphaDustEE prior: ", IntToLogical(AlphaDustEE_prior_switch)
  print *, "AlphaDustEE mean: ", AlphaDustEE_prior_mean
  print *, "AlphaDustEE prior: ", AlphaDustEE_prior_sigma

  print *, "AlphaDustTE prior: ", IntToLogical(AlphaDustTE_prior_switch)
  print *, "AlphaDustTE mean: ", AlphaDustTE_prior_mean
  print *, "AlphaDustTE prior: ", AlphaDustTE_prior_sigma

  ! Hand off to SPT3G_EETE_InitData
  call this%SPT3G_EETE_InitData(bandpower_filename, covariance_filename, beam_covariance_filename, cal_covariance_filename, window_folder, nu_eff_filename)

end subroutine SPT3G_EETE_ReadIni

subroutine SPT3G_EETE_InitData(this, bandpower_filename, covariance_filename, beam_covariance_filename, cal_covariance_filename, window_folder, nu_eff_filename)
  class(TSPT3G_EETE_Likelihood) :: this
  
  character(LEN=:), allocatable :: bandpower_filename
  real(mcp), allocatable :: full_bandpowers_with_ell(:,:)

  character(LEN=:), allocatable :: covariance_filename
  real(mcp), allocatable :: full_covariance_matrix(:,:)

  character(LEN=:), allocatable :: beam_covariance_filename
  real(mcp), allocatable :: full_beam_covariance_matrix(:,:)

  character(LEN=:), allocatable :: cal_covariance_filename
  real(mcp), allocatable :: full_cal_covariance_matrix(:,:)

  character(LEN=:), allocatable :: window_folder
  real(mcp), allocatable :: current_window(:,:)

  character(LEN=:), allocatable :: nu_eff_filename
  character(LEN=:), allocatable :: current_spec_str, current_map_1_str, current_map_2_str, current_field, current_freq_1, current_freq_2
  real(mcp), allocatable :: full_nu_eff(:,:)
  real(mcp) current_nu_eff_1, current_nu_eff_2

  integer, dimension(3) :: freqs_use

  integer i,j,ix,ix_1,ix_2,selected_spec_1,selected_spec_2,full_spec_1,full_spec_2, i0, i1, j0, j1, cal_mod
  integer cov_check

  double precision, allocatable :: dummy_cov(:,:)

  ! Check that the bin selection is valid
  if (this%N_b < 1) then
    print *, "SPT-3G Y1 EE/TE: Selected an invalid number of bandpowers: ", this%N_b
    stop
  end if

  ! Read in bandpowers
  ! N_b x N_s matrix
  allocate(full_bandpowers_with_ell(N_b_0,N_s_0+1)) ! TT and ells included
  allocate(this%full_bandpowers(N_b_0,N_s_0))
  call File%ReadTextMatrix(bandpower_filename, full_bandpowers_with_ell)
  this%full_bandpowers = full_bandpowers_with_ell(:,2:) ! Remove the first column, which is the bin index

  ! Get the band power indices of the spectra we want to fit
  allocate(spectra_to_fit_bandpower_indices(this%N_s))
  spectra_to_fit_bandpower_indices = -1
  do i=1,this%N_s
    do j=1,full_bandpower_list%Count
      if (spectra_to_fit_list%Item(i) .eq. full_bandpower_list%Item(j)) then
        spectra_to_fit_bandpower_indices(i) = j
      end if
    end do
    if (spectra_to_fit_bandpower_indices(i) .eq. -1) then
      print *, "SPT-3G Y1 EE/TE: Could not find the following spectrum in the bandpower file: ", spectra_to_fit_list%Item(i)
      stop
    end if
  end do

  ! Read in the covariance matrix
  ! N_b*N_s x N_b*N_s matrix, spectra-major order
  allocate(full_covariance_matrix(N_b_0*N_s_0,N_b_0*N_s_0))
  call File%ReadTextMatrix(covariance_filename, full_covariance_matrix)

  ! Check that all the requested spectra are in the covariance
  do i=1,this%N_s
    cov_check = -1
    do j=1,full_covariance_list%Count
      if (spectra_to_fit_list%Item(i) .eq. full_covariance_list%Item(j)) then
        cov_check = 1
      end if
    end do
    if (cov_check .eq. -1) then
      print *, "SPT-3G Y1 EE/TE: Could not find the following spectrum in the covariance file: ", spectra_to_fit_list%Item(i)
      stop
    end if
  end do

  ! Crop the covariance down to the spectra we want to fit
  allocate(this%bdp_covariance(this%N_b*this%N_s,this%N_b*this%N_s))
  this%bdp_covariance = 0!-9999
  do selected_spec_1=1,this%N_s
    do selected_spec_2=1,this%N_s
      do full_spec_1=1,full_covariance_list%Count
        do full_spec_2=1,full_covariance_list%Count
          if (spectra_to_fit_list%Item(selected_spec_1) .eq. full_covariance_list%Item(full_spec_1)) then
            if (spectra_to_fit_list%Item(selected_spec_2) .eq. full_covariance_list%Item(full_spec_2)) then
              i0 = this%bin_min+(full_spec_1-1)*N_b_0
              i1 = i0+this%N_b-1
              j0 = this%bin_min+(full_spec_2-1)*N_b_0
              j1 = j0+this%N_b-1
              this%bdp_covariance(1+(selected_spec_1-1)*this%N_b:selected_spec_1*this%N_b,1+(selected_spec_2-1)*this%N_b:selected_spec_2*this%N_b)= &
                full_covariance_matrix(i0:i1,j0:j1)
            end if
          end if
        end do
      end do
    end do
  end do

  ! Do the same procedure for the beam covariance
  allocate(full_beam_covariance_matrix(N_b_0*N_s_0,N_b_0*N_s_0))
  call File%ReadTextMatrix(beam_covariance_filename, full_beam_covariance_matrix)
  do i=1,this%N_s
    cov_check = -1
    do j=1,full_beam_covariance_list%Count
      if (spectra_to_fit_list%Item(i) .eq. full_beam_covariance_list%Item(j)) then
        cov_check = 1
      end if
    end do
    if (cov_check .eq. -1) then
      print *, "SPT-3G Y1 EE/TE: Could not find the following spectrum in the beam covariance file: ", spectra_to_fit_list%Item(i)
      stop
    end if
  end do

  allocate(this%beam_covariance(this%N_b*this%N_s,this%N_b*this%N_s))
  do selected_spec_1=1,this%N_s
    do selected_spec_2=1,this%N_s
      do full_spec_1=1,full_beam_covariance_list%Count
        do full_spec_2=1,full_beam_covariance_list%Count
          if (spectra_to_fit_list%Item(selected_spec_1) .eq. full_covariance_list%Item(full_spec_1)) then
            if (spectra_to_fit_list%Item(selected_spec_2) .eq. full_covariance_list%Item(full_spec_2)) then
              i0 = this%bin_min+(full_spec_1-1)*N_b_0
              i1 = i0+this%N_b-1
              j0 = this%bin_min+(full_spec_2-1)*N_b_0
              j1 = j0+this%N_b-1
              this%beam_covariance(1+(selected_spec_1-1)*this%N_b:selected_spec_1*this%N_b,1+(selected_spec_2-1)*this%N_b:selected_spec_2*this%N_b)= &
                full_beam_covariance_matrix(i0:i1,j0:j1)
            end if
          end if
        end do
      end do
    end do
  end do
  this%beam_covariance = this%beam_covariance * BeamCovScaling ! Allow for some arbitrary scaling

  ! Go through spectra and determine whether we need to crop any part of the cal covariance out, get inverse covariance
  ! The order of the cal covariance must always be T90, T150, T220, E90, E150, E220
  ! Also grab the indices of spectra for the Tcal and Ecal in DataParams
  allocate(spectra_to_fit_tcal_indices(this%N_s,2))
  allocate(spectra_to_fit_ecal_indices(this%N_s,2))
  cal_rows_use = 0
  do i=1,this%N_s
    ! Get frequencies of spectrum
    current_spec_str = spectra_to_fit_list%Item(i)
    ix = scan(current_spec_str, "x")
    current_map_1_str = current_spec_str(1:ix-1)
    current_map_2_str = current_spec_str(ix+1:)
    ix_1 = scan(current_map_1_str, "_")
    ix_2 = scan(current_map_2_str, "_")
    current_field = trim(current_map_1_str(ix_1+1:)) // trim(current_map_2_str(ix_2+1:))
    current_freq_1 = current_map_1_str(1:ix_1-1)
    current_freq_2 = current_map_2_str(1:ix_2-1)
    cal_mod = LogicalToInt(( trim(current_map_1_str(ix_1+1:)) .ne. trim(current_map_2_str(ix_2+1:)) )) * 3
    if ((current_freq_1 .eq. "90") .or. (current_freq_2 .eq. "90")) then
      cal_rows_use(4) = 1
      cal_rows_use(4-cal_mod) = 1
      freqs_use(1) = 1
    end if
    if ((current_freq_1 .eq. "150") .or. (current_freq_2 .eq. "150")) then
      cal_rows_use(5) = 1
      cal_rows_use(5-cal_mod) = 1
      freqs_use(2) = 1
    end if
    if ((current_freq_1 .eq. "220") .or. (current_freq_2 .eq. "220")) then
      cal_rows_use(6) = 1
      cal_rows_use(6-cal_mod) = 1
      freqs_use(3) = 1
    end if
    spectra_to_fit_tcal_indices(i,1) = LogicalToInt((current_freq_1 .eq. "90"))*iMapTcal90 +&
      LogicalToInt((current_freq_1 .eq. "150"))*iMapTcal150 +&
      LogicalToInt((current_freq_1 .eq. "220"))*iMapTcal220
    spectra_to_fit_tcal_indices(i,2) = LogicalToInt((current_freq_2 .eq. "90"))*iMapTcal90 +& 
      LogicalToInt((current_freq_2 .eq. "150"))*iMapTcal150 +&
      LogicalToInt((current_freq_2 .eq. "220"))*iMapTcal220
    spectra_to_fit_ecal_indices(i,1) = LogicalToInt((current_freq_1 .eq. "90"))*iMapEcal90 +&
      LogicalToInt((current_freq_1 .eq. "150"))*iMapEcal150 +&
      LogicalToInt((current_freq_1 .eq. "220"))*iMapEcal220
    spectra_to_fit_ecal_indices(i,2) = LogicalToInt((current_freq_2 .eq. "90"))*iMapEcal90 +&
      LogicalToInt((current_freq_2 .eq. "150"))*iMapEcal150 +&
      LogicalToInt((current_freq_2 .eq. "220"))*iMapEcal220
  end do
  this%N_freq = sum(freqs_use)

  allocate(full_cal_covariance_matrix(N_freq_0*2,N_freq_0*2))
  allocate(this%cal_inv_covariance(sum(cal_rows_use), sum(cal_rows_use)))
  call file%ReadTextMatrix(cal_covariance_filename, full_cal_covariance_matrix)
 
  i0 = 1
  do i=1,N_freq_0*2
    if (cal_rows_use(i) .eq. 1) then
      j0 = 1
      do j=1,N_freq_0*2
        if (cal_rows_use(j) .eq. 1) then
          this%cal_inv_covariance(i0,j0) = full_cal_covariance_matrix(i,j)
          j0 = j0 + 1
        end if
      end do
      i0 = i0 + 1
    end if
  end do
  call Matrix_Inverse(this%cal_inv_covariance)

  ! Go through window folder and read them all in
  ! N_l x N_s matrix for each bin
  ! Read in all bins and spectra - selection and cropping is made by in LogLike function
  allocate(this%full_windows(N_b_0,1+SPT3G_windows_lmax-SPT3G_windows_lmin,N_s_0))
  allocate(current_window(1+SPT3G_windows_lmax-SPT3G_windows_lmin,N_s_0+1)) ! +1 more specs than needed because of ells in the file
  do i=1,N_b_0
    call file%ReadTextMatrix(trim(window_folder)//trim(numcat("window_",i))//trim(".txt"), current_window)
    this%full_windows(i,:,:) = current_window(:,2:)
  end do

  ! Get the window function indices of the spectra we want to fit
  allocate(spectra_to_fit_window_indices(this%N_s))
  spectra_to_fit_window_indices = -1
  do i=1,this%N_s
    do j=1,full_window_list%Count
      if (spectra_to_fit_list%Item(i) .eq. full_window_list%Item(j)) then
        spectra_to_fit_window_indices(i) = j
      end if
    end do
    if (spectra_to_fit_window_indices(i) .eq. -1) then
      print *, "SPT-3G Y1 EE/TE: Could not find the following spectrum in the window file: ", spectra_to_fit_list%Item(i)
      stop
    end if
  end do

  ! Read in the central frequencies of all bands
  ! N_freq_0 x 1 Matrix
  allocate(full_nu_eff(N_freq_0,1))
  call File%ReadTextMatrix(nu_eff_filename, full_nu_eff)

  ! Loop over all spectra and grab their effective frequencies
  ! N_s x 2 Matrix
  allocate(this%spectra_to_fit_nu_eff(this%N_s,2))
  this%spectra_to_fit_nu_eff = -1

  do i=1,this%N_s
    ! Get frequencies of spectrum
    current_spec_str = spectra_to_fit_list%Item(i)
    ix = scan(current_spec_str, "x")
    current_map_1_str = current_spec_str(1:ix-1)
    current_map_2_str = current_spec_str(ix+1:)
    ix_1 = scan(current_map_1_str, "_")
    ix_2 = scan(current_map_2_str, "_")
    current_freq_1 = current_map_1_str(1:ix_1-1)
    current_freq_2 = current_map_2_str(1:ix_2-1)
    ! Identify effective frequencies
    current_nu_eff_1=-1
    current_nu_eff_2=-1
    do j=1,N_freq_0
      if (current_freq_1 .eq. nu_eff_list%Item(j)) then
        current_nu_eff_1 = full_nu_eff(j,1)
      end if
      if (current_freq_2 .eq. nu_eff_list%Item(j)) then
        current_nu_eff_2 = full_nu_eff(j,1)
      end if
    end do
    ! Check that frequencies were found
    if ( (current_nu_eff_1 .eq. -1) .or. (current_nu_eff_2 .eq. -1) ) then
      print *, "SPT-3G Y1 EE/TE: Could not find the effective frequencies for the following spectrum in the effective frequencies file: ", current_spec_str
      stop
    end if
    ! Store effective frequencies
    this%spectra_to_fit_nu_eff(i,1) = current_nu_eff_1
    this%spectra_to_fit_nu_eff(i,2) = current_nu_eff_2
  end do

  ! Get the indices of the poisson amplitudes of spectra we want to fit
  if (use_simple_poisson_foregrounds) then
    allocate(spectra_to_fit_poisson_indices(this%N_s))
    do i=1,this%N_s
      ! Get frequencies of spectrum
      current_spec_str = spectra_to_fit_list%Item(i)
      ix = scan(current_spec_str, "x")
      current_map_1_str = current_spec_str(1:ix-1)
      current_map_2_str = current_spec_str(ix+1:)
      ix_1 = scan(current_map_1_str, "_")
      ix_2 = scan(current_map_2_str, "_")
      current_freq_1 = current_map_1_str(1:ix_1-1)
      current_freq_2 = current_map_2_str(1:ix_2-1)
      if ((current_freq_1 .eq. "90") .AND. (current_freq_2 .eq. "90")) then
          spectra_to_fit_poisson_indices(i) = iPoisson_90x90
      else if ((current_freq_1 .eq. "90") .AND. (current_freq_2 .eq. "150")) then
          spectra_to_fit_poisson_indices(i) = iPoisson_90x150
      else if ((current_freq_1 .eq. "90") .AND. (current_freq_2 .eq. "220")) then
          spectra_to_fit_poisson_indices(i) = iPoisson_90x220
      else if ((current_freq_1 .eq. "150") .AND. (current_freq_2 .eq. "150")) then
          spectra_to_fit_poisson_indices(i) = iPoisson_150x150
      else if ((current_freq_1 .eq. "150") .AND. (current_freq_2 .eq. "220")) then
          spectra_to_fit_poisson_indices(i) = iPoisson_150x220
      else if ((current_freq_1 .eq. "220") .AND. (current_freq_2 .eq. "220")) then
          spectra_to_fit_poisson_indices(i) = iPoisson_220x220
      else
          print *, "SPT-3G Y1 EE/TE: Could not find a poisson amplitude parameter for the following spectrum in the parameter file: ", spectra_to_fit_list%Item(i)
      end if
    end do
  end if

  ! There are probably better ways to do this,
  ! Confirm successful initialisation
  print *, "SPT-3G Y1 EE/TE: Likelihood successfully initialised!"
  print *, "Fitting spectra:"
  do i=1,this%N_s
    print *, spectra_to_fit_list%Item(i)
  end do
  print *, "Over the bin range: ", this%bin_min, trim(" - "), this%bin_max
  print *, "--------------------"

end subroutine SPT3G_EETE_InitData

function SPT3G_EETE_LogLike(this, CMB, Theory, DataParams) result(SPT_LogLike)
  implicit none
  class(TSPT3G_EETE_Likelihood) :: this
  class(CMBParams) :: CMB
  class(TCosmoTheoryPredictions), target :: Theory
  real(mcp) :: DataParams(:)
  real(mcp) :: SPT_LogLike, SPT_PriorLogLike, detcov

  integer i, j, i_spec, ix, ix_1, ix_2
  character(LEN=:), allocatable :: current_spec_str, current_map_1_str, current_map_2_str, current_field, current_freq_1, current_freq_2
  
  real(mcp), dimension(SPT3G_windows_lmax+1) :: current_Dl_theory_unbinned_CMB_only ! Length is a buffer for derivatives
  real(mcp), dimension(SPT3G_windows_lmin:SPT3G_windows_lmax) :: current_Dl_theory_unbinned
  real(mcp), dimension(SPT3G_windows_lmax+1) :: Cl_derivative 
  real(mcp), dimension(this%N_b) :: current_Dl_theory_binned
  real(mcp), dimension(this%N_b*this%N_s) :: full_Dl_data_theory_binned
  real(mcp), dimension(this%N_b*this%N_s) :: Dl_data_theory_difference
  real(mcp), dimension(this%N_b*this%N_s,this%N_b*this%N_s) :: covariance, cov_copy

  real(mcp), dimension(SPT3G_windows_lmin:SPT3G_windows_lmax) :: super_sample_lensing
  real(mcp), dimension(SPT3G_windows_lmin:SPT3G_windows_lmax) :: aberration_correction
  real(mcp), dimension(SPT3G_windows_lmin:SPT3G_windows_lmax) :: simple_poisson
  real(mcp), dimension(SPT3G_windows_lmin:SPT3G_windows_lmax) :: radio_galaxies
  real(mcp), dimension(SPT3G_windows_lmin:SPT3G_windows_lmax) :: dsfg
  real(mcp), dimension(SPT3G_windows_lmin:SPT3G_windows_lmax) :: dust
  real(mcp) :: calibration
  
  real(mcp), dimension(sum(cal_rows_use)) :: cal_vec

  integer te_poisson_switch
  integer current_iRadio, current_iAlphaRadio
  integer current_iDSFG, current_iBetaDSFG
  integer current_iDust, current_iAlphaDust, current_iBetaDust

  real(mcp), dimension(SPT3G_windows_lmax+1) :: ells

  real*4, dimension(7) :: save_arr ! Array used to print out the foreground components
  integer fid, l
  
  fid = 33 ! Channel to write through

  ! Grab ells helper
  do i=1,SPT3G_windows_lmax+1
   ells(i) = i
  end do

  ! Loop over all spectra and get frequency scaling
  ! Only do this here if the scaling depends on the nuisance parameters
  ! Calculate the scaling for each map
  ! Add an array in the ini that reduces the requested spectra to what maps we have to do
  ! Loop over those maps and calculate the scaling
  ! Save in a way easily accessible in the spectra loop below (maybe just 2 N_s long freq_1/2 arrays)

  ! Loop over all spectra and add foregrounds
  ! Everything is in Dl space unless otherwise explicitly mentioned
  do i_spec=1,this%N_s

    ! Identify spectrum (get field and frequencies)
    current_spec_str = spectra_to_fit_list%Item(i_spec)
    ix = scan(current_spec_str, "x")
    current_map_1_str = current_spec_str(1:ix-1)
    current_map_2_str = current_spec_str(ix+1:)
    ix_1 = scan(current_map_1_str, "_")
    ix_2 = scan(current_map_2_str, "_")
    current_field = trim(current_map_1_str(ix_1+1:)) // trim(current_map_2_str(ix_2+1:))
    current_freq_1 = current_map_1_str(1:ix_1-1)
    current_freq_2 = current_map_2_str(1:ix_2-1)

    ! Grab field specifics
    ! Theory CMB and foregrounds that are handled differently
    if (current_field .eq. "EE") then
      call Theory%ClArray(current_Dl_theory_unbinned_CMB_only,CL_E,CL_E)

      te_poisson_switch = 1

      current_iRadio = iDl_Radio_EE_150 ! If simple foregrounds are used these point to the wrong params, but they never get used
      current_iAlphaRadio = iAlphaRadio_EE
      current_iDSFG = iDl_DSFG_EE_150
      current_iBetaDSFG = iBetaDSFG_EE
      
      current_iDust = iADust_EE_150
      current_iAlphaDust = iAlphaDust_EE
      current_iBetaDust = iBetaDust_EE

      ! Calibration for EE: 1/(Ecal_1*Ecal_2) since we matched the EE spectrum to Planck's
      calibration = 1/( DataParams(spectra_to_fit_ecal_indices(i_spec,1))*DataParams(spectra_to_fit_ecal_indices(i_spec,2)) )

    else if (current_field .eq. "TE") then
      call Theory%ClArray(current_Dl_theory_unbinned_CMB_only,CL_T,CL_E)
      
      te_poisson_switch = 0

      current_iRadio = iDl_Radio_TE_150
      current_iAlphaRadio = iAlphaRadio_TE
      current_iDSFG = iDl_DSFG_TE_150
      current_iBetaDSFG = iBetaDSFG_TE

      current_iDust = iADust_TE_150
      current_iAlphaDust = iAlphaDust_TE
      current_iBetaDust = iBetaDust_TE
    
      ! Calibration for TE: 0.5*(1/(Tcal_1*Ecal_2) + 1/(Tcal_2*Ecal_1))
      calibration = 1 /( DataParams(spectra_to_fit_tcal_indices(i_spec,1))*DataParams(spectra_to_fit_ecal_indices(i_spec,2)) )+&
        1 /( DataParams(spectra_to_fit_tcal_indices(i_spec,2))*DataParams(spectra_to_fit_ecal_indices(i_spec,1)) )
      calibration = calibration * 0.5

    else
      print *, "Cannot recognise field in spectrum: ", current_spec_str
      stop
    end if

    ! Grab Cl derivative for convenience
    Cl_derivative = current_Dl_theory_unbinned_CMB_only*2*PI/(ells*(ells+1)) ! Convert to Cl
    Cl_derivative(2:SPT3G_windows_lmax) = 0.5*( Cl_derivative(3:SPT3G_windows_lmax+1)&
      - Cl_derivative(1:SPT3G_windows_lmax-1) ) ! Take derivative
    Cl_derivative(1) = Cl_derivative(2) ! Handle endpoints approximately
    Cl_derivative(SPT3G_windows_lmax+1) = Cl_derivative(SPT3G_windows_lmax)
    if (CosmoSettings%lmax_computed_cl .LT. SPT3G_windows_lmax) then
      Cl_derivative(CosmoSettings%lmax_computed_cl) = 0.75*Cl_derivative(CosmoSettings%lmax_computed_cl-1)&
        + 0.25*Cl_derivative(CosmoSettings%lmax_computed_cl+2) ! Smooth spike from Boltzmann solver to look-up table in CAMB
      Cl_derivative(CosmoSettings%lmax_computed_cl+1) = 0.75*Cl_derivative(CosmoSettings%lmax_computed_cl+2)&
        + 0.25*Cl_derivative(CosmoSettings%lmax_computed_cl-1)
    end if

    ! Add CMB
    current_Dl_theory_unbinned = current_Dl_theory_unbinned_CMB_only(SPT3G_windows_lmin:SPT3G_windows_lmax)

    ! Add super sample lensing
    ! (In Cl space) SSL = -k/l^2 d/dln(l) (l^2Cl) = -k(l*dCl/dl + 2Cl)
    super_sample_lensing = ells(SPT3G_windows_lmin:SPT3G_windows_lmax)*Cl_derivative(SPT3G_windows_lmin:SPT3G_windows_lmax) ! l*dCl/dl
    super_sample_lensing = super_sample_lensing * ells(SPT3G_windows_lmin:SPT3G_windows_lmax)&
      *(ells(SPT3G_windows_lmin:SPT3G_windows_lmax)+1)/(2*PI) ! Convert this part to Dl space already
    super_sample_lensing = super_sample_lensing + 2*current_Dl_theory_unbinned_CMB_only(SPT3G_windows_lmin:SPT3G_windows_lmax) ! 2Dl
    super_sample_lensing = -1 * super_sample_lensing * DataParams(iKappa) ! -kappa
    super_sample_lensing = super_sample_lensing * this%ssl_switch ! Switch on/off
    current_Dl_theory_unbinned = current_Dl_theory_unbinned + super_sample_lensing

    ! Aberration correction
    ! AC = beta*l(l+1)dCl/dln(l)/(2pi)
    ! Note that the CosmoMC internal aberration correction and the SPTpol Henning likelihood differ
    ! CosmoMC uses dCl/dl, Henning et al dDl/dl
    ! In fact, CosmoMC is correct: https://journals-aps-org.eu1.proxy.openathens.net/prd/pdf/10.1103/PhysRevD.89.023003
    aberration_correction = -1*this%aberration_coefficient*Cl_derivative(SPT3G_windows_lmin:SPT3G_windows_lmax)&
      *ells(SPT3G_windows_lmin:SPT3G_windows_lmax)&
      *ells(SPT3G_windows_lmin:SPT3G_windows_lmax)&
      *(ells(SPT3G_windows_lmin:SPT3G_windows_lmax)+1)/(2*PI)
    current_Dl_theory_unbinned = current_Dl_theory_unbinned + aberration_correction

    ! Check the poisson foreground model
    if (use_simple_poisson_foregrounds) then
    
        ! Simple poisson foregrounds
        ! This is any poisson power. Meant to describe both radio galaxies and DSFG. By giving each frequency combination an amplitude
        ! to play with this gives complete freedom to the data
        simple_poisson = te_poisson_switch*ells(SPT3G_windows_lmin:SPT3G_windows_lmax)*(ells(SPT3G_windows_lmin:SPT3G_windows_lmax)+1)&
          *DataParams(spectra_to_fit_poisson_indices(i_spec))/(3000*3001)
        simple_poisson = simple_poisson * this%poisson_switch ! Switch on/off
        current_Dl_theory_unbinned = current_Dl_theory_unbinned + simple_poisson

    else

        ! Unresovled radio galaxies
        ! Amplitude is defined at l=3000 in Dl space, so we canvert to Cl power first and then to Dl
        radio_galaxies = ells(SPT3G_windows_lmin:SPT3G_windows_lmax)*(ells(SPT3G_windows_lmin:SPT3G_windows_lmax)+1)*DataParams(current_iRadio)/(3000*3001)
        radio_galaxies = radio_galaxies * RadioGalFreqScaling(DataParams(current_iAlphaRadio),nu_0_radio,this%spectra_to_fit_nu_eff(i_spec,1))&
          *RadioGalFreqScaling(DataParams(current_iAlphaRadio),nu_0_radio,this%spectra_to_fit_nu_eff(i_spec,2))
        radio_galaxies = radio_galaxies * this%poisson_switch ! Switch on/off
        current_Dl_theory_unbinned = current_Dl_theory_unbinned + radio_galaxies

        ! Dusty star-forming galaxies
        ! Same thing as unresolved radio galaxies, but have a different frequency scaling
        dsfg = ells(SPT3G_windows_lmin:SPT3G_windows_lmax)*(ells(SPT3G_windows_lmin:SPT3G_windows_lmax)+1)*DataParams(current_iDSFG)/(3000*3001)
        dsfg = dsfg * DSFGFreqScaling(DataParams(current_iBetaDSFG),DataParams(iTDSFG),nu_0_dsfg,this%spectra_to_fit_nu_eff(i_spec,1))&
          *DSFGFreqScaling(DataParams(current_iBetaDSFG),DataParams(iTDSFG),nu_0_dsfg,this%spectra_to_fit_nu_eff(i_spec,2))
        dsfg = dsfg * this%poisson_switch ! Switch on/off
        current_Dl_theory_unbinned = current_Dl_theory_unbinned + dsfg

    end if

    ! Polarised galactic dust
    dust = DataParams(current_iDust)*(ells(SPT3G_windows_lmin:SPT3G_windows_lmax)/80d0)**(DataParams(current_iAlphaDust)+2.0d0)
    dust = dust * DustFreqScaling(DataParams(current_iBetaDust),DataParams(iTDust),nu_0_dust,this%spectra_to_fit_nu_eff(i_spec,1))&
      *DustFreqScaling(DataParams(current_iBetaDust),DataParams(iTDust),nu_0_dust,this%spectra_to_fit_nu_eff(i_spec,2))
    dust = dust * this%dust_switch ! Switch on/off
    current_Dl_theory_unbinned = current_Dl_theory_unbinned + dust

    ! Scale by calibration
    current_Dl_theory_unbinned = current_Dl_theory_unbinned * calibration

    ! Bin
    ! Select the right bins from the window function in this step
    ! (Difference to SPTpol likelihood due to the changed window function format (already transposed)
    call dgemv('N',this%N_b,SPT3G_windows_lmax-SPT3G_windows_lmin+1,1.0d0,&
      this%full_windows(this%bin_min:this%bin_max,:,spectra_to_fit_window_indices(i_spec)),&
      this%N_b,current_Dl_theory_unbinned,1,0d0,current_Dl_theory_binned,1)
    full_Dl_data_theory_binned((i_spec-1)*this%N_b+1:i_spec*this%N_b) = current_Dl_theory_binned

    ! Write smooth and binned theory out
    if (print_spectrum) then
      save_arr=0
      call OpenWriteBinaryFile(trim("likelihood_tests/binned_theory_")//trim(current_spec_str),fid,4_8 * 2)
      do l=1,this%N_b
        save_arr(1)=l
        save_arr(2)=current_Dl_theory_binned(l)
        write(fid,rec=l) save_arr(1:2)
      enddo
      close(fid)
    end if

    if (print_spectrum) then
      save_arr=0
      call OpenWriteBinaryFile(trim("likelihood_tests/unbinned_theory_")//trim(current_spec_str),fid,4_8 * 2)
      do l=SPT3G_windows_lmin,SPT3G_windows_lmax
        save_arr(1)=l
        save_arr(2)=current_Dl_theory_unbinned_CMB_only(l)
        save_arr(3)=aberration_correction(l)
        save_arr(4)=super_sample_lensing(l)
        if (use_simple_poisson_foregrounds) then
          save_arr(6) = simple_poisson(l)
        else
          save_arr(5)=radio_galaxies(l)
          save_arr(6)=dsfg(l)
        end if
        save_arr(7)=dust(l)
        write(fid,rec=l-SPT3G_windows_lmin+1) save_arr(1:7)
      enddo
      close(fid)
    end if

    ! Take the difference to the measured bandpowers and preapre long (data-model) vector
    ! Crop data bandpowers to the right bins in this step too
    Dl_data_theory_difference((i_spec-1)*this%N_b+1:i_spec*this%N_b) = &
      current_Dl_theory_binned - this%full_bandpowers(this%bin_min:this%bin_max,spectra_to_fit_bandpower_indices(i_spec))

    ! Write binned model-data out
    if (print_spectrum) then
      save_arr=0
      call OpenWriteBinaryFile(trim("likelihood_tests/delta_data_model_")//trim(current_spec_str),fid,4_8 * 2)
      do l=this%bin_min,this%bin_max
        save_arr(1)=l
        save_arr(2)=current_Dl_theory_binned(l-this%bin_min+1)
        save_arr(3)=this%full_bandpowers(l,spectra_to_fit_bandpower_indices(i_spec))
        save_arr(4)=current_Dl_theory_binned(l-this%bin_min+1) - this%full_bandpowers(l,spectra_to_fit_bandpower_indices(i_spec))
        save_arr(5)=Dl_data_theory_difference((i_spec-1)*this%N_b+1+l-this%bin_min)
        write(fid,rec=l-this%bin_min+1) save_arr(1:5)
      enddo
      close(fid)
    end if

  end do

  ! Construct the full covariance matrix
  covariance=0
  do i=1,this%N_b*this%N_s
    do j=1,this%N_b*this%N_s
      covariance(i,j)=this%beam_covariance(i,j)*full_Dl_data_theory_binned(i)*full_Dl_data_theory_binned(j)
    end do
  end do
  covariance = covariance + this%bdp_covariance
  cov_copy = covariance

  ! Calculate data likelihood
  SPT_LogLike = Matrix_GaussianLogLikeDouble(covariance, Dl_data_theory_difference)

  ! Get the contribution from the determinant part
  detcov = Matrix_GaussianLogLikeDouble(cov_copy, Dl_data_theory_difference*0)

  ! Get cal vector
  ix=1
  do i=1,N_freq_0*2
    if (cal_rows_use(i) .eq. 1) then
      cal_vec(ix) = DataParams(iMapTcal90+i-1)! Cheeky, relies on the order of calibration parameters in DataParams
      ix = ix + 1
    end if
  end do
  cal_vec = log(cal_vec)

  ! Calculate prior likelihood
  SPT_PriorLogLike = 0.0d0
  SPT_PriorLogLike = SPT_PriorLogLike + Cal_prior_switch*0.5d0*Matrix_QuadForm(this%cal_inv_covariance,cal_vec)
  SPT_PriorLogLike = SPT_PriorLogLike + Kappa_prior_switch*0.5d0*((DataParams(iKappa) - Kappa_prior_mean)/Kappa_prior_sigma)**2
  SPT_PriorLogLike = SPT_PriorLogLike + AlphaDustEE_prior_switch*0.5d0*((DataParams(iAlphaDust_EE) - AlphaDustEE_prior_mean)/AlphaDustEE_prior_sigma)**2
  SPT_PriorLogLike = SPT_PriorLogLike + AlphaDustTE_prior_switch*0.5d0*((DataParams(iAlphaDust_TE) - AlphaDustTE_prior_mean)/AlphaDustTE_prior_sigma)**2

  ! Print out likelihood contributions
  if (feedback > 1) then
    print *, "SPT-3G Y1 EE/TE: Data LogLike: ", SPT_LogLike
    print *, "SPT-3G Y1 EE/TE: Data LogLike Detcov: ", detcov
    print *, "SPT-3G Y1 EE/TE: Prior LogLike: ", SPT_PriorLogLike
  end if

  ! Add data and prior likelihoods and return
  SPT_LogLike = SPT_LogLike + SPT_PriorLogLike 

end function SPT3G_EETE_LogLike


! Dust Frequency Scaling
function DustFreqScaling(beta,Tdust,nu0,nu_eff) result(fdust)
  real(mcp), intent(in) :: beta
  real(mcp), intent(in) :: Tdust
  real(mcp), intent(in) :: nu_eff
  real(mcp), intent(in) :: nu0 ! Pivot frequency
  real(mcp) :: fdust

  fdust = (nu_eff/nu0)**beta
  fdust = fdust*Bnu(nu_eff,nu0,Tdust)/dBdT(nu_eff,nu0,T_CMB)

end function DustFreqScaling


! Poisson Radio Galaxies Frequency scaling
function RadioGalFreqScaling(alpha,nu0,nu_eff) result(fradiogal)
  real(mcp), intent(in) :: nu_eff
  real(mcp), intent(in) :: alpha
  real(mcp), intent(in) :: nu0 ! Pivot frequency
  real(mcp) :: fradiogal
  
  fradiogal = (nu_eff/nu0)**alpha
  fradiogal = fradiogal/dBdT(nu_eff,nu0,T_CMB)

end function RadioGalFreqScaling


! Dusty Star-Forming Galaxies Frequency scaling
function DSFGFreqScaling(beta,Tdust,nu0,nu_eff) result(fdsfg)
  real(mcp), intent(in) :: nu_eff
  real(mcp), intent(in) :: beta
  real(mcp), intent(in) :: Tdust
  real(mcp), intent(in) :: nu0 ! Pivot frequency
  real(mcp)  :: fdsfg

  fdsfg = (nu_eff/nu0)**beta
  fdsfg = fdsfg*Bnu(nu_eff,nu0,Tdust)/dBdT(nu_eff,nu0,T_CMB)


end function DSFGFreqScaling


! Planck function normalised to 1 at nu0
function Bnu(nu,nu0,T)
  real(mcp) Bnu,nu,nu0,T

  Bnu = (nu/nu0)**3
  Bnu = Bnu * (exp( Ghz_Kelvin*nu0/T)-1d0) / (exp( Ghz_Kelvin*nu/T)-1d0)

end function Bnu


! Derivative of Planck function normalised to 1 at nu0
function dBdT(nu,nu0,T)
  real(mcp) dBdT,dBdT0,nu,nu0,T,x,x0

  x0 = Ghz_Kelvin*nu0/T
  x = Ghz_Kelvin*nu/T

  dBdT0 = x0**4 * exp(x0) / (exp(x0)-1)**2
  dBdT =  x**4 * exp(x) / (exp(x)-1)**2

  dBdT = dBdT/dBdT0

end function dBdT


! Convert boolean flag from to 0/1 and the other way
! Need to do it this way to protect for different versions of fortran (in some .true. is -1)
function LogicalToInt(log_var) result(int_var)
  logical, intent(in) :: log_var
  integer :: int_var
  int_var = 0
  if (log_var .eqv. .true.) then
    int_var = 1
  end if
end function LogicalToInt


function IntToLogical(int_var) result(log_var)
  integer, intent(in) :: int_var
  logical :: log_var
  log_var = .false.
  if (int_var .eq. 1) then
    log_var = .true.
  end if
end function IntToLogical


! Used for writing out spectra when testing the likelihood
subroutine OpenWriteBinaryFile(aname,aunit,record_length)
  character(LEN=*), intent(IN) :: aname
  integer, intent(in) :: aunit
  integer*8,intent(in) :: record_length
  open(unit=aunit,file=aname,form='unformatted',status='replace',access='direct',recl=record_length, err=500)
  return
500 call MpiStop('File not able to be written to: '//trim(aname))
end subroutine OpenWriteBinaryFile


end module CMB_SPT3G_EETE_2020
